(define-module (ide minibuffer)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)

  #:use-module (ide console)
  #:use-module (ide command)
  #:use-module (ide introspection)
  #:use-module (ide interface)
  #:use-module (ide keyboard)
  #:export (global-quit
            kill-minibuffer-or-quit
            execute-extended-command
            command-completer
            ide-procedure-completer
            minibuffer-complete
            minibuffer-done
            minibuffer-return))

(define (minibuffer-complete)
  (format (current-error-port) "<complete>\n")
  #t)

(define (minibuffer-return)
  (format (current-error-port) "<minibuffer-return>\n")
  #t)

(define (minibuffer-done)
  (set! current-key-map global-key-map)
  (minibuffer-return-focus))

;; FIXME: Misuse console for completions popup
(define %save-console #f)
(define (hide-completions)
  (when %save-console
    (console-clear)
    (console-insert %save-console)
    (set! %save-console #f)))

(define (save-console)
  (unless %save-console
    (set! %save-console (console-string))))

(define (show-completions lst)
  (save-console)
  (console-clear)
  (console-insert "Possible completions:\n")
  (let ((output-port (make-soft-port
                      (vector (const #f) (lambda (s) (console-insert s)) #f (const #f) #f) "w")))
    (with-output-to-port output-port
      (cut display-tabulated lst))))

(define* (display-tabulated lst
                            #:key
                            (terminal-width 80)
                            (column-gap 2))
  "Display the list of string LST in as many columns as needed given
TERMINAL-WIDTH.  Use COLUMN-GAP spaces between two subsequent columns."
  (define len (length lst))
  (define column-width
    ;; The width of a column.  Assume all the columns have the same width
    ;; (GNU ls is smarter than that.)
    (+ column-gap (reduce max 0 (map string-length lst))))
  (define columns
    (max 1
         (quotient terminal-width column-width)))
  (define pad
    (if (zero? (modulo len columns))
        0
        columns))
  (define items-per-column
    (quotient (+ len pad) columns))
  (define items (list->vector lst))

  (let loop ((indexes (unfold (cut >= <> columns)
                              (cut * <> items-per-column)
                              1+
                              0)))
    (unless (>= (first indexes) items-per-column)
      (for-each (lambda (index)
                  (let ((item (if (< index len)
                                  (vector-ref items index)
                                  "")))
                    (display (string-pad-right item column-width))))
                indexes)
      (newline)
      (loop (map 1+ indexes)))))

(define (common-prefix lst)
  (and (pair? lst)
       (let ((count (length lst)))
         (if (= count 1) (car lst)
             (let loop ((prefix (substring (car lst) 0 1)))
               (let ((matches (filter (cut string-prefix? prefix <>) lst)))
                 (if (< (length matches) count) (and (not (string-null? prefix))
                                                     (string-drop-right prefix 1))
                     (if (= (string-length prefix)
                            (string-length (car lst))) prefix
                      (loop (substring (car lst) 0 (1+ (string-length prefix))))))))))))

(define (command-completer)
  (let* ((name (minibuffer-string))
         (commands (delete-duplicates (map car (all-commands)) eq?))
         (matches (map symbol->string commands))
         (matches (filter (cut string-prefix? name <>) matches))
         (common (common-prefix matches)))
    (if (<=(length matches) 1) (hide-completions)
        (show-completions matches))
    (when common (minibuffer-insert common))))

(define (ide-procedure-completer)
  (let* ((name (minibuffer-string))
         (commands (delete-duplicates (map car (all-ide-procedures)) eq?))
         (matches (map symbol->string commands))
         (matches (filter (cut string-prefix? name <>) matches))
         (common (common-prefix matches)))
    (if (<=(length matches) 1) (hide-completions)
        (show-completions matches))
    (when common (minibuffer-insert common))))

(define (execute-extended-command)
  (set! global-key-map minibuffer-key-map)
  (set! current-key-map minibuffer-key-map)
  (set! minibuffer-complete command-completer)
  (set! minibuffer-return
        (lambda _
          (let* ((string (minibuffer-string))
                 (proc (catch #t (cut primitive-eval (string->symbol string)) (const #f))))
            (hide-completions)
            (minibuffer-clear)
            (minibuffer-done)
            (if (procedure? proc) (proc)
                (message (format #f "error: no such procedure: ~a\n" string))))))
  (minibuffer-prefix "M-x "))

(define (directory? file-name)
  (let ((file-name (expand-file-name file-name)))
    (and (access? file-name F_OK)
         (eq? 'directory (stat:type (stat file-name))))))

(define (canonize-file-name name)
  (if (directory? (expand-file-name name))
      (cond ((string-suffix? "/" name) name)
            (else (string-append name "/")))
      name))

(define (expand-file-name file-name)
  (let ((home (getenv "HOME")))
    (cond ((string-prefix? "~/" file-name) (string-append home (substring file-name 1)))
          ((equal? file-name "~") file-name)
          ((string-prefix? "~" file-name)
           (let* ((slash (string-index file-name #\/))
                  (user (if slash (substring file-name 1 slash)
                            (substring file-name 1))))
             (string-append (dirname home) "/" user (if slash (substring file-name slash) ""))))
          ((string-prefix? "~" file-name) file-name)
          (else file-name))))

(define (contract-file-name file-name)
  (let ((home (getenv "HOME")))
    (cond ((string=? file-name home) "~/")
          ((string-prefix? home file-name)
           (string-append "~" (substring file-name (string-length home))))
          (else file-name))))

(define (sort-file-names lst)
  (sort lst string<))

(define (files-in-dir dirname)
  (let* ((dir (opendir (expand-file-name dirname)))
         (file-names (unfold eof-object? identity (lambda _ (readdir dir)) (readdir dir))))
    (closedir dir)
    (sort-file-names file-names)))

(define (dirname- name)
  (let ((home (getenv "HOME")))
    (cond ((string=? name "~") (dirname home))
          ((or (string=? name "~/")
               (and (string-prefix? "~/" name)
                    (not (string-index (substring name 2) #\/)))) "~/")
          ((and (string-prefix? "~" name)
                (not (string-index name #\/))) (dirname home))
          ((directory? (expand-file-name name)) (canonize-file-name name))
          (else (dirname name)))))

(define (file-name-completer)
  (let ((file-name (minibuffer-string)))
    (cond ((string-contains file-name "//") => (lambda (p) (minibuffer-insert (substring file-name (1+ p)))))
          ((string-contains file-name "/~/") => (lambda (p) (minibuffer-insert (substring file-name (1+ p)))))
          ((string-contains file-name "/~") => (lambda (p) (minibuffer-insert (string-append (substring file-name (1+ p)) "/"))))
          ((string-contains file-name "/./") => (lambda (p) (minibuffer-insert (string-replace file-name "" p (+ p 2)))))
          ((string-contains file-name "/..") => (lambda (p)
                                                  (let* ((prefix (substring file-name 0 p))
                                                         (prefix (dirname prefix)))
                                                    (minibuffer-insert (string-append prefix (substring file-name (+ p 3))))))))
    (let* ((file-name (minibuffer-string))
           (base (if (directory? file-name) ""
                     (basename file-name)))
           (dir (if (and (directory? file-name)
                         (string-suffix? "/" file-name)) file-name
                         (dirname- file-name)))
           (files (files-in-dir (dirname- file-name)))
           (matches (filter (compose (cut string-prefix? base <>) basename) files))
           (common (common-prefix matches)))
      (if (<= (length matches) 1) (hide-completions)
          (show-completions (if (pair? matches) matches files)))
      (when common
        (minibuffer-insert (canonize-file-name (string-append (canonize-file-name dir) common)))))))

(define %default-directory (canonize-file-name (getcwd)))
(define-interactive (find-file)
  (let ((key-map (filter (negate (compose (cut equal? "SPACE" <>) car)) minibuffer-key-map)))
    (set! global-key-map key-map)
    (set! current-key-map key-map))
  (set! minibuffer-complete file-name-completer)
  (set! minibuffer-return
        (lambda _
          (let* ((file-name (minibuffer-string))
                 (file-name (expand-file-name file-name)))
            (set! %default-directory (canonize-file-name (dirname file-name)))
            (hide-completions)
            (minibuffer-clear)
            (minibuffer-done)
            (visit-file file-name))))
  (minibuffer-prefix "Find file: ")
  (minibuffer-insert (contract-file-name %default-directory)))

(define (global-quit)
  (unselect)
  (minibuffer-prefix "")
  (message "")
  (minibuffer-quit))

(define (kill-minibuffer-or-quit)  
  (hide-completions)
  (global-quit))

(define (error-msg-handler key subr msg args . rest)
  (with-output-to-string
    (lambda _
      (display-error #f (current-output-port) subr msg args rest))))

(define* (with-error-msg-handler thunk #:optional (error-msg-handler error-msg-handler))
  (catch #t thunk error-msg-handler))

(define-interactive (eval-expression)
  (let ((key-map (filter (negate (compose (cut equal? "SPACE" <>) car)) minibuffer-key-map)))
    (set! global-key-map key-map)
    (set! current-key-map key-map))
  (set! minibuffer-return
        (lambda _
          (let* ((string (minibuffer-string))
                 (sexp (with-error-msg-handler (lambda _ (with-input-from-string string read))))
                 (result (with-error-msg-handler (lambda _ (primitive-eval sexp)))))
            (hide-completions)
            (minibuffer-clear)
            (message (format #f "~s" result))
            (minibuffer-done)
            result)))
  (minibuffer-prefix "Eval: "))

(define-interactive (shell-command)
  (let ((key-map (filter (negate (compose (cut equal? "SPACE" <>) car)) minibuffer-key-map)))
    (set! global-key-map key-map)
    (set! current-key-map key-map))
  (set! minibuffer-return
        (lambda _
          (let ((command (minibuffer-string)))
            (minibuffer-clear)
            (let ((result (run-shell-command command)))
              (minibuffer-done)
              result))))
  (minibuffer-prefix "Shell command: "))

(define-interactive (async-shell-command)
  (let ((key-map (filter (negate (compose (cut equal? "SPACE" <>) car)) minibuffer-key-map)))
    (set! global-key-map key-map)
    (set! current-key-map key-map))
  (set! minibuffer-return
        (lambda _
          (let ((command (minibuffer-string)))
            (minibuffer-clear)
            (let ((result (run-shell-command-async command)))
              (minibuffer-done)
              result))))
  (minibuffer-prefix "Shell command: "))
